/*
 * gbiff2 program
 * Copyright (C) 2000-2002 Nicolas Rougier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <stdlib.h>
#include <gtk/gtk.h>
#include <glade/glade.h>
#include <glade/glade-build.h>

#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>

#include <gtk/gtkwindow.h>
#include <gtk/gtkprivate.h>
#include <gtk/gtkrc.h>
#include <gtk/gtksignal.h>
#include <gtk/gtkwindow.h>
#include <gtk/gtkbindings.h>
#include <gtk/gtkmain.h>

#include "support.h"


/********************************************************************************
 * Function: Automatic signal connection for libglade, using gmodule
 * Input   : /
 * Output  : /
 * Return  : /
 * Note    : This function is necessary for passing user_data with autoconnect
 ********************************************************************************/
void MyGladeXMLConnectFunc (const gchar *handler_name,
                            GObject *object,
                            const gchar *signal_name,
                            const gchar *signal_data,
                            GObject *connect_object,
                            gboolean after,
                            gpointer user_data) {
  GModule *symbols = 0;
  GtkSignalFunc func;
 
  if (!symbols) {
    if (!g_module_supported()) {
      g_error ("GladeXMLConnectFunc requires working gmodule");
      exit (1);
    }
    symbols = g_module_open (0, G_MODULE_BIND_MASK);
  }
 
  if (!g_module_symbol(symbols, handler_name, (gpointer *) &func))
    g_warning(_("Could not find signal handler '%s'."), handler_name);
  else
    g_signal_connect (object, signal_name, GTK_SIGNAL_FUNC(func), user_data);
}

/********************************************************************************
 * Function: Get the specified widget from the xml tree
 * Input   : A biff structure and the name of the widget
 * Output  : /
 * Return  : Widget or 0
 * Note    : /
 ********************************************************************************/
GtkWidget *__widget (Biff *biff, gchar *wname) {
  GtkWidget *widget = glade_xml_get_widget (biff->xml, wname);
  if (widget == 0)
    g_warning (_("Could not find the specified widget (\"%s\") within the xml tree"), wname);
  return widget;
}


/********************************************************************************
 * Function: locale/filename/utf8 conversion functions
 * Input   : A locale/filename/utf8 string
 * Output  : /
 * Return  : A locale/filename/utf8 string
 * Note    : See function name for input/return type
 ********************************************************************************/
gchar *filename_from_utf8 (const gchar *string) {
  static gchar *buffer = 0;

  if (buffer)
    g_free (buffer);
  buffer = g_filename_from_utf8 (string,-1,0,0,0);
  return buffer;
}

gchar *filename_to_utf8 (gchar *string) {
  static gchar *buffer = 0;

  if (buffer)
    g_free (buffer);
  buffer = g_filename_to_utf8 (string,-1,0,0,0);
  return buffer;
}

gchar *locale_from_utf8 (const gchar *string) {
  static gchar *buffer = 0;

  if (buffer)
    g_free (buffer);
  buffer = g_locale_from_utf8 (string,-1,0,0,0);
  return buffer;
}

gchar *locale_to_utf8 (gchar *string) {
  static gchar *buffer = 0;

  if (buffer)
    g_free (buffer);
  buffer = g_locale_to_utf8 (string,-1,0,0,0);
  return buffer;
}


/********************************************************************************
 * Function: Get a pango font and handle memory allocations
 * Input   : A string describing a font
 * Output  : /
 * Return  : A pango font description
 * Note    : /
 ********************************************************************************/
PangoFontDescription *getfont (gchar *fontname) {
  static PangoFontDescription *font = 0;

  if (font != 0)
    pango_font_description_free (font);
  font = pango_font_description_from_string (fontname);
  return font;
}




/********************************************************************************
 * The code below is extracted from gtk 2.0 gtkwindow.c
 * ----------------------------------------------------
 * The goal is to be able to parse geometry correctly, that is, taking window
 * decorations into acount
 * The only change compared to original version is that we use
 * gdk_window_get_frame_extents to get window real size and place the window.
 ********************************************************************************/


/*
  Derived from XParseGeometry() in XFree86  

  Copyright 1985, 1986, 1987,1998  The Open Group

  All Rights Reserved.

  The above copyright notice and this permission notice shall be included
  in all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
  IN NO EVENT SHALL THE OPEN GROUP BE LIABLE FOR ANY CLAIM, DAMAGES OR
  OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
  ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
  OTHER DEALINGS IN THE SOFTWARE.

  Except as contained in this notice, the name of The Open Group shall
  not be used in advertising or otherwise to promote the sale, use or
  other dealings in this Software without prior written authorization
  from The Open Group.
*/



typedef struct
{
  GList     *icon_list;
  GdkPixmap *icon_pixmap;
  GdkPixmap *icon_mask;
  guint      realized : 1;
  guint      using_default_icon : 1;
  guint      using_parent_icon : 1;
} GtkWindowIconInfo;

typedef struct {
  GdkGeometry    geometry; /* Last set of geometry hints we set */
  GdkWindowHints flags;
  GdkRectangle   configure_request;
} GtkWindowLastGeometryInfo;

struct _GtkWindowGeometryInfo
{
  /* Properties that the app has set on the window
   */
  GdkGeometry    geometry;	/* Geometry hints */
  GdkWindowHints mask;
  GtkWidget     *widget;	/* subwidget to which hints apply */
  /* from last gtk_window_resize () - if > 0, indicates that
   * we should resize to this size.
   */
  gint           resize_width;  
  gint           resize_height;

  /* From last gtk_window_move () prior to mapping -
   * only used if initial_pos_set
   */
  gint           initial_x;
  gint           initial_y;
  
  /* Default size - used only the FIRST time we map a window,
   * only if > 0.
   */
  gint           default_width; 
  gint           default_height;
  /* whether to use initial_x, initial_y */
  guint          initial_pos_set : 1;
  /* CENTER_ALWAYS or other position constraint changed since
   * we sent the last configure request.
   */
  guint          position_constraints_changed : 1;

  /* if true, default_width, height come from gtk_window_parse_geometry,
   * and thus should be multiplied by the increments and affect the
   * geometry widget only
   */
  guint          default_is_geometry : 1;
  
  GtkWindowLastGeometryInfo last;
};

/*
 *    XParseGeometry parses strings of the form
 *   "=<width>x<height>{+-}<xoffset>{+-}<yoffset>", where
 *   width, height, xoffset, and yoffset are unsigned integers.
 *   Example:  "=80x24+300-49"
 *   The equal sign is optional.
 *   It returns a bitmask that indicates which of the four values
 *   were actually found in the string.  For each value found,
 *   the corresponding argument is updated;  for each value
 *   not found, the corresponding argument is left unchanged. 
 */

/* The following code is from Xlib, and is minimally modified, so we
 * can track any upstream changes if required.  Don't change this
 * code. Or if you do, put in a huge comment marking which thing
 * changed.
 */

static int
read_int (gchar   *string,
          gchar  **next)
{
  int result = 0;
  int sign = 1;
  
  if (*string == '+')
    string++;
  else if (*string == '-')
    {
      string++;
      sign = -1;
    }

  for (; (*string >= '0') && (*string <= '9'); string++)
    {
      result = (result * 10) + (*string - '0');
    }

  *next = string;

  if (sign >= 0)
    return (result);
  else
    return (-result);
}

/* Try not to reformat/modify, so we can compare/sync with X sources */
int
gtk_XParseGeometry (const char   *string,
                    int          *x,
                    int          *y,
                    unsigned int *width,   
                    unsigned int *height)  
{
  int mask = NoValue;
  char *strind;
  unsigned int tempWidth, tempHeight;
  int tempX, tempY;
  char *nextCharacter;

  /* These initializations are just to silence gcc */
  tempWidth = 0;
  tempHeight = 0;
  tempX = 0;
  tempY = 0;
  
  if ( (string == NULL) || (*string == '\0')) return(mask);
  if (*string == '=')
    string++;  /* ignore possible '=' at beg of geometry spec */

  strind = (char *)string;
  if (*strind != '+' && *strind != '-' && *strind != 'x') {
    tempWidth = read_int(strind, &nextCharacter);
    if (strind == nextCharacter) 
      return (0);
    strind = nextCharacter;
    mask |= WidthValue;
  }

  if (*strind == 'x' || *strind == 'X') {	
    strind++;
    tempHeight = read_int(strind, &nextCharacter);
    if (strind == nextCharacter)
      return (0);
    strind = nextCharacter;
    mask |= HeightValue;
  }

  if ((*strind == '+') || (*strind == '-')) {
    if (*strind == '-') {
      strind++;
      tempX = -read_int(strind, &nextCharacter);
      if (strind == nextCharacter)
        return (0);
      strind = nextCharacter;
      mask |= XNegative;

    }
    else
      {	strind++;
      tempX = read_int(strind, &nextCharacter);
      if (strind == nextCharacter)
        return(0);
      strind = nextCharacter;
      }
    mask |= XValue;
    if ((*strind == '+') || (*strind == '-')) {
      if (*strind == '-') {
        strind++;
        tempY = -read_int(strind, &nextCharacter);
        if (strind == nextCharacter)
          return(0);
        strind = nextCharacter;
        mask |= YNegative;

      }
      else
        {
          strind++;
          tempY = read_int(strind, &nextCharacter);
          if (strind == nextCharacter)
            return(0);
          strind = nextCharacter;
        }
      mask |= YValue;
    }
  }
	
  /* If strind isn't at the end of the string the it's an invalid
		geometry specification. */

  if (*strind != '\0') return (0);

  if (mask & XValue)
    *x = tempX;
  if (mask & YValue)
    *y = tempY;
  if (mask & WidthValue)
    *width = tempWidth;
  if (mask & HeightValue)
    *height = tempHeight;

  return (mask);
}

/**
 * gtk_window_get_destroy_with_parent:
 * @window: a #GtkWindow
 * 
 * Returns whether the window will be destroyed with its transient parent. See
 * gtk_window_set_destroy_with_parent ().
 *
 * Return value: %TRUE if the window will be destroyed with its transient parent.
 **/
gboolean
gtk_window_get_destroy_with_parent (GtkWindow *window)
{
  g_return_val_if_fail (GTK_IS_WINDOW (window), FALSE);

  return window->destroy_with_parent;
}

static GtkWindowGeometryInfo*
gtk_window_get_geometry_info (GtkWindow *window,
			      gboolean   create)
{
  GtkWindowGeometryInfo *info;

  info = window->geometry_info;
  if (!info && create)
    {
      info = g_new0 (GtkWindowGeometryInfo, 1);

      info->default_width = -1;
      info->default_height = -1;
      info->resize_width = -1;
      info->resize_height = -1;
      info->initial_x = 0;
      info->initial_y = 0;
      info->initial_pos_set = FALSE;
      info->default_is_geometry = FALSE;
      info->position_constraints_changed = FALSE;
      info->last.configure_request.x = 0;
      info->last.configure_request.y = 0;
      info->last.configure_request.width = -1;
      info->last.configure_request.height = -1;
      info->widget = NULL;
      info->mask = 0;
      window->geometry_info = info;
    }

  return info;
}

/********************************************************************************
 * Function: Parses a standard X Window System geometry string
 *            and (only) move window accordingly
 * Input   : A strign describing a font
 * Output  : /
 * Return  : A pango font description
 * Note    : /
 ********************************************************************************/
gboolean my_gtk_window_parse_geometry (GtkWindow *window, const gchar *geometry) {
  gint result, x, y;
  guint w, h;
  GdkGravity grav;
  gboolean size_set, pos_set;
  GdkRectangle rect;
  
  g_return_val_if_fail (GTK_IS_WINDOW (window), FALSE);
  g_return_val_if_fail (geometry != NULL, FALSE);
  
  result = gtk_XParseGeometry (geometry, &x, &y, &w, &h);
  if ((result & WidthValue) == 0 || w < 0)
    w = -1;
  if ((result & HeightValue) == 0 || h < 0)
    h = -1;

  size_set = FALSE;
  if ((result & WidthValue) || (result & HeightValue))
    {
//      gtk_window_set_default_size_internal (window, TRUE, w, TRUE, h, TRUE);
      size_set = TRUE;
    }

  gdk_window_get_frame_extents (GTK_WIDGET(window)->window, &rect);
  w = rect.width;
  h = rect.height;

  grav = GDK_GRAVITY_NORTH_WEST;

  if ((result & XNegative) && (result & YNegative))
    grav = GDK_GRAVITY_SOUTH_EAST;
  else if (result & XNegative)
    grav = GDK_GRAVITY_NORTH_EAST;
  else if (result & YNegative)
    grav = GDK_GRAVITY_SOUTH_WEST;

  if ((result & XValue) == 0)
    x = 0;

  if ((result & YValue) == 0)
    y = 0;

  if (grav == GDK_GRAVITY_SOUTH_WEST ||
      grav == GDK_GRAVITY_SOUTH_EAST)
    y = gdk_screen_height () - h + y;

  if (grav == GDK_GRAVITY_SOUTH_EAST ||
      grav == GDK_GRAVITY_NORTH_EAST)
    x = gdk_screen_width () - w + x;

  /* we don't let you put a window offscreen; maybe some people would
   * prefer to be able to, but it's kind of a bogus thing to do.
   */
  if (y < 0)
    y = 0;

  if (x < 0)
    x = 0;

  pos_set = FALSE;
  if ((result & XValue) || (result & YValue))
    {
      gtk_window_set_gravity (window, grav);
      gtk_window_move (window, x, y);
      pos_set = TRUE;
    }

  if (size_set || pos_set)
    {
      /* Set USSize, USPosition hints */
      GtkWindowGeometryInfo *info;
      info = gtk_window_get_geometry_info (window, TRUE);
      if (pos_set)
        info->mask |= GDK_HINT_USER_POS;
      if (size_set)
        info->mask |= GDK_HINT_USER_SIZE;
    }
  
  return result != 0;
}
